from jieba import Tokenizer, strdecode
from typing import List
import threading


class ParallelJiebaTokenizer(Tokenizer):

    def __init__(self, dictionary=None):
        super().__init__(dictionary=dictionary)
        super().initialize(dictionary=dictionary)

    def strdecode(self, sentence):
        return strdecode(sentence)

    def lcut(
            self,
            sentence: str,
            cut_all: bool = False,
            HMM: bool = True,
        ) -> List[str]:
        return super().lcut(sentence, cut_all=cut_all, HMM=HMM)

    def __getstate__(self):
        d = super().__dict__.copy()
        if 'lock' in d:
            del d['lock']
        return d

    def __setstate__(self, d):
        self.lock = threading.RLock()
        self.__dict__.update(d)



from unittest import TestCase

from .. import NltkCustomJiebaTokenizer


class NltkCustomJiebaTokenizerTestCase(TestCase):

    def setUp(self):
        self.tokenizer = NltkCustomJiebaTokenizer()

    def test_default_lcut(self):
        tokenizer = NltkCustomJiebaTokenizer()
        self.assertEqual(
            tokenizer.lcut("我想喝Mr.S珍珠奶茶半糖大杯"),
            self.tokenizer.lcut(
                "我想喝Mr.S珍珠奶茶半糖大杯",
                punct=True,
                cut_all=False,
                HMM=False,
            ),
        )
        self.assertEqual(
            ["我", "想", "喝", "Mr", ".", "S", "珍珠奶茶", "半", "糖", "大", "杯"],
            tokenizer.lcut("我想喝Mr.S珍珠奶茶半糖大杯"),
        )

    def test_lcut_punct_false(self):
        tokenizer = NltkCustomJiebaTokenizer(punct=False)
        self.assertEqual(
            tokenizer.lcut("How's it going today, Mr.Smith?"),
            self.tokenizer.lcut(
                "How's it going today, Mr.Smith?",
                punct=False,
            ),
        )
        self.assertEqual(
            ["How", "'s", "it", "going", "today", ",", "Mr.Smith", "?", ],
            tokenizer.lcut("How's it going today, Mr.Smith?"),
        )

    def test_lcut_HMM_false(self):
        tokenizer = NltkCustomJiebaTokenizer(HMM=False)
        self.assertEqual(
            tokenizer.lcut("我想喝Mr.S珍珠奶茶半糖大杯"),
            self.tokenizer.lcut("我想喝Mr.S珍珠奶茶半糖大杯", HMM=False),
        )
        self.assertEqual(
            ["我", "想", "喝", "Mr", ".", "S", "珍珠奶茶", "半", "糖", "大", "杯"],
            tokenizer.lcut("我想喝Mr.S珍珠奶茶半糖大杯"),
        )

    def test_lcut_sentences_parallel(self):
        human_tokened_sentences = [
            ["我", "想", "喝", "珍珠奶茶", "半", "糖", "大", "杯"],
            ["i", "would", "like", "to", "drink", "pearl", "milk", "tea", "."],
            ["_ya_", "我", "想", "喝", "_num_", "杯", "_test_", "飲", "料", "_end_"],
            ["我", "想", "買", "12", "張", "(", "abc", ")", "股"],
            ["我", "想", "買", "12", "張", "abc", "股"],
        ]
        origin_sentences = [
            "我想喝珍珠奶茶半糖大杯",
            "i would like to drink pearl milk tea.",
            "_ya_ 我想喝 _num_ 杯 _test_ 飲料 _end_",
            "我想買12張(abc)股",
            "我想買 12 張 abc 股",
        ]
        tokenized_sentences = self.tokenizer.lcut_sentences(
            origin_sentences,
            num_jobs=8,
            HMM=False,
        )
        self.assertListEqual(human_tokened_sentences, tokenized_sentences)



from typing import List
from .base_tokenizer import BaseTokenizer


class AddWords(object):

    def __init__(
            self,
            tokenizer: BaseTokenizer,
            extra_words: List[str] = None,
        ) -> None:
        if extra_words is None:
            extra_words = []
        self.extra_words = extra_words
        self.tokenizer = tokenizer

    def __enter__(self):
        if hasattr(self.tokenizer, "add_word_idempotent"):
            self.existed_tokens = {}
            for word in self.extra_words:
                self.existed_tokens.update(
                    self.tokenizer.add_word_idempotent(word=word),
                )
        return self.tokenizer

    def __exit__(self, type, value, traceback):
        if hasattr(self.tokenizer, "del_word_idempotent"):
            for word in self.extra_words:
                self.tokenizer.del_word_idempotent(
                    word=word,
                    existed_tokens=self.existed_tokens,
                )



from typing import List
import nltk
nltk.download('punkt')  # noqa
import functools
from nltk.tokenize import WordPunctTokenizer, word_tokenize

from .base_tokenizer import BaseTokenizer


class NltkTokenizer(BaseTokenizer):

    def __init__(
            self,
            punct: bool = None,
        ):
        self.WPtokenizer = WordPunctTokenizer()
        self.punct = punct

    def lcut(
            self,
            sentence,
            punct: bool = None,
            **kwargs  # noqa
        ):
        if punct is not None:
            self.punct = punct
        if self.punct is True:
            return(self.WPtokenizer.tokenize(sentence))
        else:
            return(word_tokenize(sentence))

    def lcut_sentences(
            self,
            sentences: List[str],
            num_jobs: int = 8,
            punct: bool = True,
        ) -> List[List[str]]:
        from multiprocessing import cpu_count, Pool
        if num_jobs is None:
            num_jobs = cpu_count()
        if self.punct is not None:
            punct = self.punct
        with Pool(num_jobs) as pool:
            results = pool.map(
                functools.partial(
                    self.lcut,
                    punct=punct,
                ),
                sentences,
            )
        return results



from itertools import permutations


int_with_digits = ["_{}int_".format(num) for num in range(1, 13)]
float_with_digits = [
    "_{}float{}_".format(pair[0], pair[1]) for pair in list(
        permutations(list(range(1, 13)), 2)
    ) if pair[1] < 5
]

RESERVED_TOKENS = [
    "_int_",
    "_float_",
    "_num_",
] + int_with_digits + float_with_digits


from .purewords_tokenizer import PureWordsTokenizer  # noqa
from .parallel_jieba_tokenizer import ParallelJiebaTokenizer  # noqa
from .chinese_char_tokenizer import ChineseCharTokenizer  # noqa
from .pure_char_tokenizer import PureChineseCharTokenizer  # noqa
from .custom_jieba_tokenizer import CustomJiebaTokenizer  # noqa
from .nltk_tokenizer import NltkTokenizer  # noqa
from .nltk_custom_jieba_tokenizer import NltkCustomJiebaTokenizer  # noqa
from .add_words import AddWords  # noqa



import re
from typing import List

import purewords

from .base_tokenizer import BaseTokenizer
from .parallel_jieba_tokenizer import ParallelJiebaTokenizer


class PureWordsTokenizer(BaseTokenizer):

    def __init__(
            self,
            dict_path: str = None,
            freq_words: List[str] = None,
        ):
        self.tokenizer = ParallelJiebaTokenizer()
        if dict_path is not None:
            self.tokenizer.load_userdict(dict_path)
        if freq_words is not None:
            self._add_words(words=freq_words)
        self.specific_tokens = [
            '_url_',
            '_num_',
            '_phone_',
            '_time_',
        ]
        self._add_words(
            words=self.specific_tokens,
        )
        self.purewords_tokenizer = purewords.PureWords(
            tokenizer=self.tokenizer,
        )
        self.sub_prog_1 = re.compile(r"(?<=\s)_(\s)")
        self.sub_prog_2 = re.compile(r"\s+")

    def _add_words(
            self,
            words: List[str],
            freq: int = None,
            tag: str = None,
        ):
        for word in words:
            self.tokenizer.add_word(word=word, freq=freq, tag=tag)
            self.tokenizer.suggest_freq(segment=word, tune=True)

    def lcut(
            self,
            sentence: str,
            **kwargs  # noqa
        ) -> List[str]:
        clean_sentence = self.purewords_tokenizer.clean_sentence(sentence)
        clean_sentence = ' ' + clean_sentence + ' '
        clean_sentence = self.sub_prog_1.sub(" ", clean_sentence)
        clean_sentence = self.sub_prog_2.sub(" ", clean_sentence)
        clean_sentence = clean_sentence.strip()
        if clean_sentence == '':
            return []
        else:
            tokens = clean_sentence.split(' ')
            return tokens

    def lcut_sentences(
            self,
            sentences: List[str],
            num_jobs: int = 8,
        ) -> List[List[str]]:
        from multiprocessing import cpu_count, Pool
        if num_jobs is None:
            num_jobs = cpu_count()
        with Pool(num_jobs) as pool:
            tokenized_sentences = pool.map(self.lcut, sentences)
        return tokenized_sentences



from unittest import TestCase

from .. import NltkTokenizer


class test_NltkTokenizer(TestCase):

    def test_NltkTokenizer(self):
        tokenizer = NltkTokenizer()
        self.assertEqual(
            ['How', "'s", 'it', 'going', 'today', ',', 'Mr.Smith', '?', ],
            tokenizer.lcut("How's it going today, Mr.Smith?", punct=False),
        )
        tokenizer = NltkTokenizer(punct=False)
        self.assertEqual(
            ['How', "'s", 'it', 'going', 'today', ',', 'Mr.Smith', '?', ],
            tokenizer.lcut("How's it going today, Mr.Smith?"),
        )

    def test_NltkTokenizer_punct(self):
        tokenizer = NltkTokenizer()
        self.assertEqual(
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Mr', '.', 'Smith', '?', ],
            tokenizer.lcut("How's it going today, Mr.Smith?", punct=True),
        )
        tokenizer = NltkTokenizer(punct=True)
        self.assertEqual(
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Mr', '.', 'Smith', '?', ],
            tokenizer.lcut("How's it going today, Mr.Smith?"),
        )

    def test_NltkTokenizer_parallel_lcut(self):
        tokenizer = NltkTokenizer()
        human_tokened_sentences = [
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Mr', '.', 'Smith', '?', ],
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Mr', '.', 'John', '?', ],
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Mrs', '.', 'Watson', '?', ],
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Mrs', '.', 'Stone', '?', ],
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Dr', '.', 'Smith', '?', ],
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Dr', '.', 'John', '?', ],
            ['How', "'", 's', 'it', 'going', 'today', ',', 'Mrs', '.', 'Winslet', '?', ],
        ]
        sentences_to_token = [
            "How's it going today, Mr.Smith?",
            "How's it going today, Mr.John?",
            "How's it going today, Mrs.Watson?",
            "How's it going today, Mrs.Stone?",
            "How's it going today, Dr.Smith?",
            "How's it going today, Dr.John?",
            "How's it going today, Mrs.Winslet?",
        ]
        tokenized_sentences = tokenizer.lcut_sentences(
            sentences_to_token,
            num_jobs=4,
            punct=True,
        )
        self.assertListEqual(human_tokened_sentences, tokenized_sentences)
        tokenizer = NltkTokenizer(punct=True)
        tokenized_sentences = tokenizer.lcut_sentences(
            sentences_to_token,
            num_jobs=4,
        )
        self.assertListEqual(human_tokened_sentences, tokenized_sentences)



# -*- coding: utf-8 -*-
from unittest import TestCase

from ..chinese_char_tokenizer import (
    ChineseCharTokenizer,
)


class ChineseCharTokenizerTestCase(TestCase):

    def setUp(self):
        self.tokenizer = ChineseCharTokenizer()

    def test_chinese_char_tokenizer(self):
        test_cases = [
            ("我想要買miss.M的蛋糕", ["我", "想", "要", "買",
                               "miss", ".", "M", "的", "蛋", "糕"]),
            ("來一卡車MM chocolate", ["來", "一", "卡", "車", "MM", " ", "chocolate"]),
            ("想要買(B)股", ["想", "要", "買", "(", "B", ")", "股"]),
            ("來一杯40.77度的pink lady",
             ["來", "一", "杯", "40.77", "度", "的", "pink", " ", "lady"]),
            ("來一杯_float_度的pink lady",
             ["來", "一", "杯", "_float_", "度", "的", "pink", " ", "lady"]),
            ("_3float2_元的飲料", ["_3float2_", "元", "的", "飲", "料"]),
            ("I want _12int_ cup of bubble tea",
             ["I", " ", "want", " ", "_12int_", " ", "cup", " ", "of", " ", "bubble", " ", "tea"]),
            ("薄餡的手機號碼是0800-000-123",
             ["薄", "餡", "的", "手", "機", "號", "碼", "是", "0800", "-", "000", "-", "123"]),
            ("邊...邊緣人", ["邊", "...", "邊", "緣", "人"]),
        ]
        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.assertEqual(
                    test_case[1],
                    self.tokenizer.lcut(test_case[0]),
                )

    def test_chinese_char_tokenizer_parallel(self):
        actual_result = self.tokenizer.lcut_sentences(
            sentences=[
                "我想要買miss.M的蛋糕",
                "來一卡車MM chocolate",
                "想要買(B)股",
                "來一杯40.77度的pink lady",
                "來一杯_float_度的pink lady",
                "_3float2_元的飲料",
                "I want _12int_ cup of bubble tea",
                "薄餡的手機號碼是0800-000-123",
            ],
        )
        self.assertEqual(
            [
                ["我", "想", "要", "買", "miss", ".", "M", "的", "蛋", "糕"],
                ["來", "一", "卡", "車", "MM", " ", "chocolate"],
                ["想", "要", "買", "(", "B", ")", "股"],
                ["來", "一", "杯", "40.77", "度", "的", "pink", " ", "lady"],
                ["來", "一", "杯", "_float_", "度", "的", "pink", " ", "lady"],
                ["_3float2_", "元", "的", "飲", "料"],
                ["I", " ", "want", " ", "_12int_", " ", "cup",
                 " ", "of", " ", "bubble", " ", "tea"],
                ["薄", "餡", "的", "手", "機", "號", "碼", "是",
                 "0800", "-", "000", "-", "123"],
            ],
            actual_result,
        )



from typing import List
import functools
import re

import purewords
import string

from .base_tokenizer import BaseTokenizer


class CharTokenizerForPureWords(BaseTokenizer):

    def __init__(self):
        self.prog = re.compile(
            "[{}]+".format(re.escape(string.printable)),
        )

    def lcut(
            self,
            sentence: str,
            **kwargs  # noqa
        ):
        not_chinese_elements = self.prog.findall(sentence)
        pure_chinese_sentence = self.prog.sub("X", sentence)
        tokens = list(pure_chinese_sentence)
        index = 0
        for ti, token in enumerate(tokens):
            if token == 'X':
                tokens[ti] = not_chinese_elements[index]
                index += 1
        return tokens

    def cut(
            self,
            sentence: str,
        ) -> str:
        result = self.lcut(sentence)
        for char in result:
            yield char

    def lcut_sentences(
            self,
            sentences: List[str],
            num_jobs: int = None,
        ) -> List[List[str]]:
        from multiprocessing import cpu_count, Pool
        if num_jobs is None:
            num_jobs = cpu_count()
        with Pool(num_jobs) as pool:
            results = pool.map(
                functools.partial(
                    self.lcut,
                ),
                sentences,
            )
        return results


class PureChineseCharTokenizer(BaseTokenizer):

    def __init__(self):
        self.purechartokenizer = purewords.PureWords(
            tokenizer=CharTokenizerForPureWords(),
        )

    def lcut(
            self,
            sentence: str,
        ) -> List[str]:
        return self.purechartokenizer.clean_sentence(sentence).split(" ")

    def cut(
            self,
            sentence: str,
        ) -> str:
        result = self.lcut(sentence=sentence)
        for char in result:
            yield char



from unittest import TestCase

from .. import PureWordsTokenizer


class PurewordsTokenizerTestCase(TestCase):

    def setUp(self):
        self.tokenizer = PureWordsTokenizer()

    def test_lcut(self):
        test_cases = [
            {
                "sentence": "薄餡=柏憲=cph=cph_is_god\n讚讚讚！聯絡方式：cph@cph.tw, 0912345678",
                "answer": ['薄餡', '柏憲', 'cph', 'cph', 'is', 'god',
                           '讚', '讚', '讚', '聯絡', '方式', '_url_', '_phone_'],
            },
            {
                "sentence": "_薄餡=柏憲=cph=cph_is_god\n讚讚讚！聯絡方式：cph@cph.tw, __0912345678_",
                "answer": ['薄餡', '柏憲', 'cph', 'cph', 'is', 'god',
                           '讚', '讚', '讚', '聯絡', '方式', '_url_', '_phone_'],
            },
            {
                "sentence": "_ _ _ _ _ ",
                "answer": [],
            },
            {
                "sentence": "__ ____ _ _ _ ",
                "answer": [],
            },
        ]
        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.assertEqual(
                    test_case['answer'],
                    self.tokenizer.lcut(sentence=test_case['sentence']),
                )



from unittest import TestCase

from .. import CustomJiebaTokenizer
from ..add_words import AddWords


class CustomJiebaTokenizerTestCase(TestCase):

    def setUp(self):
        self.tokenizer = CustomJiebaTokenizer()

    def test_default_lcut(self):
        self.assertEqual(
            self.tokenizer.lcut("我想喝珍珠奶茶半糖大杯"),
            self.tokenizer.lcut(
                "我想喝珍珠奶茶半糖大杯",
                cut_all=False,
                HMM=False,
            ),
        )
        self.assertEqual(
            ["我", "想", "喝", "珍珠奶茶", "半", "糖", "大", "杯"],
            self.tokenizer.lcut("我想喝珍珠奶茶半糖大杯"),
        )

    def test_hmm_true(self):
        tokenizer = CustomJiebaTokenizer(HMM=True)
        self.assertEqual(
            tokenizer.lcut("我想喝珍珠奶茶半糖大杯"),
            self.tokenizer.lcut("我想喝珍珠奶茶半糖大杯", HMM=True),
        )
        self.assertEqual(
            ["我", "想", "喝", "珍珠奶茶", "半糖", "大杯"],
            tokenizer.lcut("我想喝珍珠奶茶半糖大杯"),
        )

    def test_cut_all_true(self):
        tokenizer = CustomJiebaTokenizer(cut_all=True)
        self.assertEqual(
            tokenizer.lcut("我想喝珍珠奶茶半糖大杯"),
            self.tokenizer.lcut("我想喝珍珠奶茶半糖大杯", cut_all=True),
        )
        self.assertEqual(
            ["我", "想", "喝", "珍珠", "珍珠奶茶", "奶茶", "半", "糖", "大杯"],
            self.tokenizer.lcut("我想喝珍珠奶茶半糖大杯", cut_all=True),
        )

    def test_lcut_with_extra_words(self):
        sentence = "我想吃珍珠奶茶鍋，加一道普羅旺斯主廚香煎烤雞排"
        extra_words = ["珍珠奶茶鍋", "普羅旺斯主廚香煎烤雞排"]
        expected_tokens = ["我", "想", "吃", "珍珠奶茶鍋",
                           "，", "加", "一道", "普羅旺斯主廚香煎烤雞排"]
        with AddWords(tokenizer=self.tokenizer, extra_words=extra_words):
            actual_tokens = self.tokenizer.lcut(sentence)
        self.assertEqual(expected_tokens, actual_tokens)

    def test_no_side_effects_after_lcut(self):
        sentence = "我想吃珍珠奶茶鍋，加一道普羅旺斯主廚香煎烤雞排"
        extra_words = ["珍珠奶茶鍋", "普羅旺斯主廚香煎烤雞排"]
        old_result = self.tokenizer.lcut(sentence)
        with AddWords(tokenizer=self.tokenizer, extra_words=extra_words):
            self.tokenizer.lcut(sentence)
        new_result = self.tokenizer.lcut(sentence)
        self.assertEqual(old_result, new_result)

    def test_lcut_sentences_parallel(self):
        human_tokened_sentences = [
            ["我", "想", "喝", "珍珠奶茶", "半", "糖", "大", "杯"],
            ["你", "想", "喝", "珍珠奶茶", "半", "糖", "小杯"],
            ["他", "想", "喝", "珍珠奶茶", "全", "糖", "大", "杯"],
            ["她", "想", "喝", "珍珠奶茶", "半", "糖", "大", "杯"],
            ["我", "想", "喝", "珍珠奶茶", "半", "糖", "中", "杯"],
        ]
        origin_sentences = [
            "我想喝珍珠奶茶半糖大杯",
            "你想喝珍珠奶茶半糖小杯",
            "他想喝珍珠奶茶全糖大杯",
            "她想喝珍珠奶茶半糖大杯",
            "我想喝珍珠奶茶半糖中杯",
        ]
        tokenized_sentences = self.tokenizer.lcut_sentences(
            origin_sentences,
            num_jobs=8,
            HMM=False,
        )
        self.assertListEqual(human_tokened_sentences, tokenized_sentences)



from typing import List
import string

import re
import functools

from .base_tokenizer import BaseTokenizer
from .nltk_tokenizer import NltkTokenizer
from .custom_jieba_tokenizer import CustomJiebaTokenizer


class NltkCustomJiebaTokenizer(BaseTokenizer):

    def __init__(
            self,
            punct: bool = None,
            cut_all: bool = None,
            HMM: bool = None,
            userdict_path: str = None,
        ) -> None:
        self.punct = punct
        self.cut_all = cut_all
        self.HMM = HMM
        self.userdict_path = userdict_path

        self.nltk_tokenizer = NltkTokenizer(
            punct=self.punct,
        )
        self.custom_jieba_tokenizer = CustomJiebaTokenizer(
            cut_all=self.cut_all,
            HMM=self.HMM,
            userdict_path=self.userdict_path,
        )
        self.prog = re.compile('[{}]+'.format(re.escape(string.printable)))

    def lcut(
            self,
            sentence: str,
            punct: bool = None,
            cut_all: bool = None,
            HMM: bool = None,
            num_jobs: int = 4,
            **kwargs  # noqa
        ) -> List[str]:
        if punct is not None:
            self.punct = punct
        if cut_all is not None:
            self.cut_all = cut_all
        if HMM is not None:
            self.HMM = HMM
        tokens = self.nltk_tokenizer.lcut(
            sentence=sentence,
            punct=self.punct,
        )
        output_tokens = []
        for token in tokens:
            eng_num = self.prog.sub('', token)
            if len(eng_num) > 0:
                output_tokens.append(
                    self.custom_jieba_tokenizer.lcut(
                        sentence=token,
                        cut_all=self.cut_all,
                        HMM=self.HMM,
                    ),
                )
            else:
                output_tokens.append([token])
        return sum(output_tokens, [])

    def lcut_sentences(
            self,
            sentences: List[str],
            punct: bool = True,
            extra_words: List[str] = None,
            cut_all: bool = False,
            HMM: bool = True,
            num_jobs: int = None,
        ) -> List[List[str]]:
        from multiprocessing import cpu_count, Pool
        if num_jobs is None:
            num_jobs = cpu_count()
        if self.punct is not None:
            punct = self.punct
        if self.cut_all is not None:
            cut_all = self.cut_all
        if self.HMM is not None:
            HMM = self.HMM
        with Pool(num_jobs) as pool:
            results = pool.map(
                functools.partial(
                    self.lcut,
                    punct=punct,
                    extra_words=extra_words,
                    cut_all=cut_all,
                    HMM=HMM,
                    num_jobs=num_jobs,
                ),
                sentences,
            )
        return results



from typing import List
import functools
import re

from .base_tokenizer import BaseTokenizer
from . import RESERVED_TOKENS


class ChineseCharTokenizer(BaseTokenizer):

    def __init__(self):
        reg_pattern = RESERVED_TOKENS + \
            ["[0-9]+\.[0-9]+"] + \
            ["[a-zA-Z]+"] + \
            ["[0-9]+"] + \
            ["\.\.\."]

        self.prog = re.compile(
            r"{}".format("|".join(reg_pattern)),
        )

    def lcut(
            self,
            sentence: str,
            **kwargs  # noqa
        ) -> List[str]:
        not_chinese_elements = self.prog.findall(sentence)
        pure_chinese_sentence = self.prog.sub("X", sentence)
        tokens = list(pure_chinese_sentence)
        index = 0
        for ti, token in enumerate(tokens):
            if token == "X":
                tokens[ti] = not_chinese_elements[index]
                index += 1
        return tokens

    def lcut_sentences(
            self,
            sentences: List[str],
            num_jobs: int = None,
        ) -> List[List[str]]:
        from multiprocessing import cpu_count, Pool
        if num_jobs is None:
            num_jobs = cpu_count()
        with Pool(num_jobs) as pool:
            results = pool.map(
                functools.partial(
                    self.lcut,
                ),
                sentences,
            )
        return results



# -*- coding: utf-8 -*-
import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

try:
    long_description = open("README.md").read()
except IOError:
    long_description = ""

about = {}
with open(os.path.join(here, "tokenizer_hub", "__version__.py")) as f:
    exec(f.read(), about)

setup(
    name="tokenizer-hub",
    version=about["__version__"],
    description="Yoctol Natural Language Tokenizer",
    url="https://github.com/Yoctol/tokenizer-hub",
    license="MIT",
    author="Solumilken",
    packages=find_packages(),
    python_requires='>=3.5',
    install_requires=[],
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)



from typing import List


class BaseTokenizer(object):

    def lcut(
            self,
            sentence: str,
        ) -> List[str]:
        """Tokenize a single sentence
        Args:
          sentence: single string representing a sentence
        Returns:
          tokenized_sentence: a list of tokens
        """
        raise NotImplementedError

    def lcut_sentences(
            self,
            sentences: List[str],
            num_jobs: int,
        ) -> List[List[str]]:
        """Tokenizing sentences in a parallel way.
        Args:
          sentences: list of sentence to be tokenized
          num_jobs: parallel processors
        Returns:
          tokenized_sentences: List[List[tokens]]
        """
        raise NotImplementedError



from typing import Dict, List

from .parallel_jieba_tokenizer import ParallelJiebaTokenizer, strdecode
from .base_tokenizer import BaseTokenizer
import functools


class CustomJiebaTokenizer(ParallelJiebaTokenizer, BaseTokenizer):

    def __init__(
            self,
            cut_all: bool = None,
            HMM: bool = None,
            userdict_path: str = None,
        ):
        super(CustomJiebaTokenizer, self).__init__()
        self.cut_all = cut_all
        self.HMM = HMM
        if userdict_path is not None:
            super(CustomJiebaTokenizer, self).load_userdict(userdict_path)

    def add_word_idempotent(
            self,
            word: str,
        ) -> Dict[str, int]:
        word = strdecode(word)
        freq = 1
        self.FREQ[word] = freq
        self.total += freq

        existed_token = {}
        for ch in range(len(word)):
            wfrag = word[:ch + 1]
            if wfrag not in self.FREQ:
                self.FREQ[wfrag] = 0
            else:
                existed_token[wfrag] = self.FREQ[wfrag]
        return existed_token

    def del_word_idempotent(
            self,
            word: str,
            existed_tokens: Dict[str, int],
        ) -> None:
        word = strdecode(word)
        self.total -= self.FREQ[word]
        del self.FREQ[word]

        for ch in range(len(word)):
            wfrag = word[:ch + 1]
            if wfrag not in existed_tokens and self.FREQ[wfrag] == 0:
                del self.FREQ[wfrag]

    def lcut(
            self,
            sentence: str,
            cut_all: bool = None,
            HMM: bool = None,
            **kwargs  # noqa
        ) -> List[str]:
        if cut_all is not None:
            self.cut_all = cut_all
        if HMM is not None:
            self.HMM = HMM
        return super().lcut(
            sentence,
            cut_all=self.cut_all,
            HMM=self.HMM,
        )

    def lcut_sentences(
            self,
            sentences: List[str],
            num_jobs: int = None,
            extra_words: List[str] = None,
            HMM: bool = True,
            cut_all: bool = False,
        )-> List[List[str]]:
        from multiprocessing import cpu_count, Pool
        if num_jobs is None:
            num_jobs = cpu_count()
        if self.cut_all is not None:
            cut_all = self.cut_all
        if self.HMM is not None:
            HMM = self.HMM
        with Pool(num_jobs) as pool:
            results = pool.map(
                functools.partial(
                    self.lcut,
                    extra_words=extra_words,
                    cut_all=cut_all,
                    HMM=HMM,
                ),
                sentences,
            )
        return results



# -*- coding: utf-8 -*-
from unittest import TestCase

from ..pure_char_tokenizer import (
    CharTokenizerForPureWords,
    PureChineseCharTokenizer,
)


class PureCharTokenizerTestCase(TestCase):

    def setUp(self):
        self.tokenizer = CharTokenizerForPureWords()
        self.pure_tokenizer = PureChineseCharTokenizer()

    def test_char_tokenizer(self):
        test_cases = [
            ("#abcAWERDFTQookirh09875462512k.-(){}[]#@!~<>=*^|_%:;?",
             ["#abcAWERDFTQookirh09875462512k.-(){}[]#@!~<>=*^|_%:;?"]),
            ("我想要買miss.M的蛋糕", ["我", "想", "要", "買", "miss.M", "的", "蛋", "糕"]),
            ("來一卡車MM chocolate", ["來", "一", "卡", "車", "MM chocolate"]),
            ("想要買(B)股", ["想", "要", "買", "(B)", "股"]),
            ("來一杯40.77度的pink lady",
             ["來", "一", "杯", "40.77", "度", "的", "pink lady"]),
            ("薄餡的手機號碼是0800-000-123",
             ["薄", "餡", "的", "手", "機", "號", "碼", "是", "0800-000-123"]),
        ]
        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.assertEqual(
                    test_case[1],
                    self.tokenizer.lcut(test_case[0]),
                )

    def test_char_tokenizer_parallel(self):
        actual_result = self.tokenizer.lcut_sentences(
            sentences=[
                "#abcAWERDFTQookirh09875462512k.-(){}[]#@!~<>=*^|_%:;?",
                "我想要買miss.M的蛋糕",
                "來一卡車MM chocolate",
                "想要買(B)股",
                "來一杯40.77度的pink lady",
                "薄餡的手機號碼是0800-000-123",
            ],
        )
        self.assertEqual(
            [
                ["#abcAWERDFTQookirh09875462512k.-(){}[]#@!~<>=*^|_%:;?"],
                ["我", "想", "要", "買", "miss.M", "的", "蛋", "糕"],
                ["來", "一", "卡", "車", "MM chocolate"],
                ["想", "要", "買", "(B)", "股"],
                ["來", "一", "杯", "40.77", "度", "的", "pink lady"],
                ["薄", "餡", "的", "手", "機", "號", "碼", "是", "0800-000-123"],
            ],
            actual_result,
        )

    def test_pure_chinese_char_tokenizer(self):
        test_cases = [
            ("我想要買miss.M的蛋糕", ['我', '想', '要',
                               '買', 'miss', 'm', '的', '蛋', '糕']),
            ("來一卡車mm chocolate<><>!!!###", [
             '來', '一', '卡', '車', 'mm', 'chocolate']),
            ("想要買(B)股", ["想", "要", "買", "b", "股"]),
            ("來一杯40.77度的pink lady",
             ["來", "一", "杯", "40", "77", "度", "的", "pink", "lady"]),
            ("薄餡的手機號碼是0800-000-123",
             ["薄", "餡", "的", "手", "機", "號", "碼", "是", "_phone_"]),
            ("豪大大想買100個豪大大雞排",
             ["豪", "大", "大", "想", "買", "_num_", "個", "豪", "大", "大", "雞", "排"]),
        ]
        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.assertEqual(
                    test_case[1],
                    self.pure_tokenizer.lcut(test_case[0]),
                )
